/*
    Project: TCGINA
    Version: V1.27
    File   : secmem.c

    Copyright (c) 2005-2009 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include <windows.h>
#include "secmem.h"

#define USE_SECURE_HEAP
#define SUPPORT_MULTITHREADING

/*
 * The definition of USE_BUFFERS allocates larger blocks for requests below
 * a certain size. The large block is evenly divided into 8*sizeof(unsigned)
 * smaller blocks where the first block is used for bookkeeping information.
 * Large blocks are only freed during sec_heap_compact() or
 * sec_heap_destroy().
 */
#define USE_BUFFERS

/*
 * The definition of USE_PAGE_SHARING allows memory blocks to belong to the
 * same memory page. A singly linked list of all allocated memory blocks is
 * maintained to defer the call of VitualUnlock() if a memory block, which
 * uses the same memory page is still in use.
 *
 * If USE_PAGE_SHARING is not defined, all allocation requests are enlarged by
 * the size of one memory page. The additional space is used as padding, and
 * the requested memory is placed between its two halfs. The padding assures
 * that no requested memory block shares a single page with another memory
 * block, and VirtualUnlock() can be called without unintentionally
 * unprotecting another memory block. The first half of the padding is used
 * for bookkeeping information.
 */
#define USE_PAGE_SHARING

//#define USE_OWN_HEAP
//#define USE_OWN_ITEM_HEAP

/*------------------------------------------------------------------------------
// Data for debugging purposes
//----------------------------------------------------------------------------*/
#ifdef SEC_DEBUG
static SEC_HEAP_INFORMATION shiStatus;
#endif

/*------------------------------------------------------------------------------
// Data for multi-threading support
//----------------------------------------------------------------------------*/

#ifdef USE_SECURE_HEAP

#ifdef SUPPORT_MULTITHREADING
static CRITICAL_SECTION csCriticalSection;
#  define USE_LOCK
#  define InitLock() InitializeCriticalSection(&csCriticalSection)
#  define DeleteLock() DeleteCriticalSection(&csCriticalSection)
#  define Lock() EnterCriticalSection(&csCriticalSection)
#  define Unlock() LeaveCriticalSection(&csCriticalSection)
#else
#  define InitLock()
#  define DeleteLock()
#  define Lock()
#  define Unlock()
#endif

/*------------------------------------------------------------------------------
// Data for secure heap
//----------------------------------------------------------------------------*/

#ifdef USE_PAGE_SHARING
#  define USE_ITEM_HEAP
#endif

struct s_ITEM;
typedef struct s_ITEM ITEM, *PITEM;
struct s_ITEM
{
    PITEM  pNextItem;
    PBYTE  pData;
    size_t uSize;
#ifdef USE_PAGE_SHARING
    size_t uFirstPage;
    size_t uLastPage;
#endif
};

static BOOL fIgnoreVirtualLockFailure = FALSE;
static DWORD  dwPageSize              = 0;
#ifdef USE_PAGE_SHARING
static size_t uPageMask               = 0;
#else
static DWORD  dwHalfPageSize          = 0;
#endif
static HANDLE hHeap                   = NULL;
#ifdef USE_ITEM_HEAP
static HANDLE hItemHeap               = NULL;
#endif
static PITEM  pFirstItem              = NULL;
static PITEM  pLastItem               = NULL;

/*------------------------------------------------------------------------------
// Data for small blocks
//----------------------------------------------------------------------------*/

#ifdef USE_BUFFERS

#define PAGE_COUNT        4
#define OPT_PAGE_SIZE     4096
#define BUFFER_ITEM_COUNT (sizeof(unsigned)*8)
#define BUFFER_ITEM_SIZE  ((OPT_PAGE_SIZE*PAGE_COUNT) / BUFFER_ITEM_COUNT)
#define BUFFER_SIZE       (BUFFER_ITEM_SIZE*BUFFER_ITEM_COUNT)
#define ALL_BUFFER_ITEMS_ARE_IN_USE (((1U<<(BUFFER_ITEM_COUNT-1U))-1U)<<1U)

struct s_BUFFER;
typedef struct s_BUFFER BUFFER, *PBUFFER;
struct s_BUFFER {
    PBUFFER  pNextBuffer;
    PBYTE    pData;
    unsigned uUsageFlags;
};

static PBUFFER pFirstBuffer = NULL;
static PBUFFER pLastBuffer  = NULL;

#endif

/*------------------------------------------------------------------------------
// Initialization and destruction
//----------------------------------------------------------------------------*/

void sec_memset(void *p, char ch, size_t uSize)
{
    volatile char *vp = (volatile char *) p;

    while (uSize) {
        *vp++ = ch;
        uSize--;
    }
}

#define wipemem(p, s) sec_memset(p, '\xff', s); sec_memset(p, '\0', s)

int sec_heap_init(void)
{
    if (hHeap != NULL) {
        return FALSE;
    }

#ifdef SEC_DEBUG
    memset(&shiStatus, '\0', sizeof(shiStatus));
#endif

    if (dwPageSize == 0) {
        SYSTEM_INFO sSysInfo;
        GetSystemInfo(&sSysInfo);
        dwPageSize = sSysInfo.dwPageSize;
        if (dwPageSize == 0) {
            dwPageSize = 4096;
        }
#ifdef USE_PAGE_SHARING
        uPageMask = ~(size_t)(dwPageSize - 1);
#else
        dwHalfPageSize = (dwPageSize + 1) / 2;
#endif
    }

    InitLock();

#ifdef USE_OWN_HEAP
    hHeap = HeapCreate(0, 0, 0);
#else
    hHeap = GetProcessHeap();
#endif
    if (hHeap == NULL) {
        DeleteLock();
        return FALSE;
    }

#ifdef USE_ITEM_HEAP
#  ifdef USE_OWN_ITEM_HEAP
    hItemHeap = HeapCreate(0, 0, 0);
#  else
    hItemHeap = GetProcessHeap();
#  endif
    if (hItemHeap == NULL) {
#  ifdef USE_OWN_HEAP
        HeapDestroy(hHeap);
#  endif
        hHeap = NULL;
        DeleteLock();
        return FALSE;
    }
#endif

    return TRUE;
}

void sec_heap_destroy(void)
{
    PITEM pItem;

    if (hHeap == NULL) {
        return;
    }

#ifdef USE_BUFFERS
    pFirstBuffer = pLastBuffer = NULL;
#endif

    for (pItem = pFirstItem; pItem != NULL; pItem = pItem->pNextItem) {
        wipemem(pItem->pData, pItem->uSize);
        VirtualUnlock(pItem->pData, pItem->uSize);
    }
    pFirstItem = pLastItem = NULL;

#ifdef USE_OWN_HEAP
    HeapDestroy(hHeap);
#endif
    hHeap = NULL;
#ifdef USE_ITEM_HEAP
#  ifdef USE_OWN_ITEM_HEAP
    HeapDestroy(hItemHeap);
#  endif
    hItemHeap = NULL;
#endif

    DeleteLock();

#ifdef SEC_DEBUG
    memset(&shiStatus, '\0', sizeof(shiStatus));
#endif
}

int sec_heap_mode(int iMode)
{
    int iPrevMode = fIgnoreVirtualLockFailure = iMode ? TRUE : FALSE;
    if (iMode >= 0) {
        fIgnoreVirtualLockFailure = iMode;
    }
    return iPrevMode;
}

/*------------------------------------------------------------------------------
// Allocation
//----------------------------------------------------------------------------*/

#ifdef USE_LOCK
static void sec_heap_free(void *pData);
#else
#  define sec_heap_free sec_free
#endif

static void * sec_heap_alloc(size_t uSize, BOOL fIgnoreVirtualLockResult)
{
    PITEM pItem;
#ifndef USE_PAGE_SHARING
    PBYTE pData;
#endif

#ifdef USE_BUFFERS
    if (uSize <= BUFFER_ITEM_SIZE) {

        PBUFFER  pBuffer;
        unsigned uUsageFlags;
        unsigned i;

        for (pBuffer = pFirstBuffer; pBuffer != NULL;
          pBuffer = pBuffer->pNextBuffer) {
            if (pBuffer->uUsageFlags != ALL_BUFFER_ITEMS_ARE_IN_USE) {
                break;
            }
        }
        if (pBuffer == NULL) {
            PBYTE pData = sec_heap_alloc(BUFFER_SIZE,
              fIgnoreVirtualLockResult);
            if (pData == NULL) {
                return NULL;
            }
            pBuffer = (PBUFFER) pData;
            pBuffer->pNextBuffer = NULL;
            pBuffer->pData = pData;
            pBuffer->uUsageFlags = 0;
            if (pLastBuffer == NULL) {
                pLastBuffer = pFirstBuffer = pBuffer;
            } else {
                pLastBuffer = pLastBuffer->pNextBuffer = pBuffer;
            }
        }
        uUsageFlags = 2;
        for (i = 1; i < BUFFER_ITEM_COUNT; i++) {
            if ((pBuffer->uUsageFlags & uUsageFlags) == 0) {
                pBuffer->uUsageFlags |= uUsageFlags;
                return pBuffer->pData + i * BUFFER_ITEM_SIZE;
            }
            uUsageFlags <<= 1;
        }

        return NULL;
    }
#else
    if (uSize == 0) {
        uSize = 1;
    }
#  if !defined(USE_LOCK) && !defined(USE_ITEM_HEAP)
    if (hItemHeap == NULL) {
        return NULL;
    }
#  endif
#endif

#ifdef USE_PAGE_SHARING
    pItem = HeapAlloc(hItemHeap, 0, sizeof(ITEM));
    if (pItem == NULL) {
        return NULL;
    }
    pItem->pData = HeapAlloc(hHeap, 0, uSize);
    if (pItem->pData == NULL) {
        HeapFree(hItemHeap, 0, pItem);
        return NULL;
    }
#else
    pData = HeapAlloc(hHeap, 0, uSize + dwPageSize);
    if (pData == NULL) {
        return NULL;
    }
    pItem = (PITEM)pData;
    pItem->pData = pData + dwHalfPageSize;
#endif

    if (!VirtualLock(pItem->pData, uSize)) {
        if (!fIgnoreVirtualLockResult) {
#  ifdef USE_PAGE_SHARING
            HeapFree(hHeap, 0, pItem->pData);
            HeapFree(hItemHeap, 0, pItem);
#  else
            HeapFree(hHeap, 0, pData);
#  endif
            return NULL;
        }
    }

    pItem->pNextItem = NULL;
    pItem->uSize = uSize;
#ifdef USE_PAGE_SHARING
    pItem->uFirstPage = ((size_t)pItem->pData) & uPageMask;
    pItem->uLastPage = (((size_t)pItem->pData) + (uSize - 1)) & uPageMask;
#endif

    if (pLastItem == NULL) {
        pLastItem = pFirstItem = pItem;
    } else {
        pLastItem = pLastItem->pNextItem = pItem;
    }

#ifdef SEC_DEBUG
    shiStatus.cAllocations++;
    if (shiStatus.cAllocations > shiStatus.cMaxAllocations) {
        shiStatus.cMaxAllocations = shiStatus.cAllocations;
    }
    shiStatus.cbAllocationSize += pItem->uSize;
    if (shiStatus.cbAllocationSize > shiStatus.cbMaxAllocationSize) {
        shiStatus.cbMaxAllocationSize = shiStatus.cbAllocationSize;
    }
#endif

    return pItem->pData;
}

void * sec_malloc(size_t uSize)
{
    void *pData;

    if (hHeap == NULL && !sec_heap_init()) {
        return NULL;
    }

    Lock();

    pData = sec_heap_alloc(uSize, FALSE);
    if (pData == NULL) {
        sec_heap_compact();
        pData = sec_heap_alloc(uSize, fIgnoreVirtualLockFailure);
    }

    Unlock();

    return pData;
}

void * sec_calloc(size_t cNum, size_t uSize)
{
    void *pData;

    if (hHeap == NULL && !sec_heap_init()) {
        return NULL;
    }

    Lock();

    uSize *= cNum;

    pData = sec_heap_alloc(uSize, FALSE);
    if (pData == NULL) {
        sec_heap_compact();
        pData = sec_heap_alloc(uSize, fIgnoreVirtualLockFailure);
    }

    if (pData) {
        memset(pData, '\0', uSize);
    }

    Unlock();

    return pData;
}

static size_t sec_heap_get_block_size(const void *pData)
{
    PITEM pItem;

    for (pItem = pFirstItem; pItem != NULL; pItem = pItem->pNextItem) {
        if (pItem->pData == pData) {
            return pItem->uSize;
        }
    }

#ifdef USE_BUFFERS
    return BUFFER_ITEM_SIZE;
#else
    return 0;
#endif
}

void * sec_realloc(void *pData, size_t uNewSize)
{
    void *pNewData;

    if (hHeap == NULL && !sec_heap_init()) {
        return NULL;
    }

    Lock();

    pNewData = sec_heap_alloc(uNewSize, FALSE);
    if (pNewData == NULL) {
        sec_heap_compact();
        pNewData = sec_heap_alloc(uNewSize, fIgnoreVirtualLockFailure);
    }

    if (pNewData != NULL && pData != NULL) {
        size_t uSize = sec_heap_get_block_size(pData);
        if (uSize > uNewSize) {
            uSize = uNewSize;
        }
        memcpy(pNewData, pData, uSize);
        sec_heap_free(pData);
    }

    Unlock();

    return pNewData;
}

/*------------------------------------------------------------------------------
// Deallocation
//----------------------------------------------------------------------------*/

#ifdef USE_LOCK
void sec_free(void *pData)
{
    if (pData == NULL) {
        return;
    }

    Lock();

    sec_heap_free(pData);

    Unlock();
}
static void sec_heap_free(void *pData)
#else
void sec_free(void *pData)
#endif
{
    PITEM pItem;
    PITEM pPrevItem;
#ifdef USE_PAGE_SHARING
    PITEM pRelatedItem;
#endif
#ifdef USE_BUFFERS
    PBUFFER pBuffer;
#endif

#ifndef USE_LOCK
    if (pData == NULL) {
        return;
    }
#endif

#ifdef USE_BUFFERS
    for (pBuffer = pFirstBuffer; pBuffer != NULL;
        pBuffer = pBuffer->pNextBuffer) {
        size_t uIdx = (((PBYTE)pData) - pBuffer->pData) /
            BUFFER_ITEM_SIZE;
        if (uIdx == 0) {
            break;
        }
        if (uIdx < BUFFER_ITEM_COUNT) {
            wipemem(pData, BUFFER_ITEM_SIZE);
            pBuffer->uUsageFlags &= ~(1<<(unsigned)uIdx);
            return;
        }
    }
#endif

    for (pItem = pFirstItem, pPrevItem = NULL; pItem != NULL;
        pItem = pItem->pNextItem) {
        if (pItem->pData == pData) {
            break;
        }
        pPrevItem = pItem;
    }

    if (pPrevItem == NULL) {
        pFirstItem = pItem->pNextItem;
    } else {
        pPrevItem->pNextItem = pItem->pNextItem;
    }
    if (pItem == pLastItem) {
        pLastItem = pPrevItem;
    }

    wipemem(pData, pItem->uSize);

#ifdef SEC_DEBUG
    shiStatus.cAllocations--;
    shiStatus.cbAllocationSize -= pItem->uSize;
#endif

#ifdef USE_PAGE_SHARING
    for (pRelatedItem = pFirstItem; pRelatedItem != NULL;
      pRelatedItem = pRelatedItem->pNextItem) {
        if (pRelatedItem->uFirstPage >= pItem->uFirstPage) {
            if (pRelatedItem->uFirstPage <= pItem->uLastPage) {
                break;
            }
        } else {
            if (pRelatedItem->uLastPage >= pItem->uFirstPage) {
                break;
            }
        }
    }
    if (pRelatedItem == NULL) {
        VirtualUnlock((PVOID)(pItem->uFirstPage),
          (pItem->uLastPage - pItem->uFirstPage + dwPageSize));
    } else {
        if (pRelatedItem->uFirstPage > pItem->uFirstPage) {
            pRelatedItem->uFirstPage = pItem->uFirstPage;
        }
        if (pRelatedItem->uLastPage < pItem->uLastPage) {
            pRelatedItem->uLastPage = pItem->uLastPage;
        }
    }

    HeapFree(hHeap, 0, pData);
    HeapFree(hItemHeap, 0, pItem);
#else
    VirtualUnlock(pItem->pData, pItem->uSize);
    HeapFree(hHeap, 0, pItem);
#endif
}

/*------------------------------------------------------------------------------
// Clean-up
//----------------------------------------------------------------------------*/

void sec_heap_compact(void)
{
#ifdef USE_BUFFERS
    PBUFFER pBuffer, pPrevBuffer, pNextBuffer;
#endif

    if (hHeap == NULL) {
        return;
    }

    Lock();

#ifdef USE_BUFFERS
    pPrevBuffer = NULL;
    for (pBuffer = pFirstBuffer; pBuffer != NULL; pBuffer = pNextBuffer) {
        pNextBuffer = pBuffer->pNextBuffer;
        if (pBuffer->uUsageFlags) {
            pPrevBuffer = pBuffer;
        } else {
            sec_heap_free(pBuffer->pData);
            if (pPrevBuffer == NULL) {
                pFirstBuffer = pNextBuffer;
            } else {
                pPrevBuffer->pNextBuffer = pNextBuffer;
            }
            if (pBuffer == pLastBuffer) {
                pLastBuffer = pPrevBuffer;
            }
        }
    }
#endif

#ifdef USE_OWN_HEAP
    HeapCompact(hHeap, 0);
#endif
#ifdef USE_ITEM_HEAP
#  ifdef USE_OWN_ITEM_HEAP
    HeapCompact(hItemHeap, 0);
#  endif
#endif

    Unlock();
}

#ifdef SEC_DEBUG
int sec_heap_status(PSEC_HEAP_INFORMATION pshi)
{
    *pshi = shiStatus;
    return TRUE;
}
#endif

#else

void sec_memset(char *p, char ch, size_t uSize) { memset(p, ch, uSize); }
BOOL sec_heap_init(void) { return FALSE; }
void sec_heap_destroy(void) { }
void * sec_malloc(size_t uSize) { return malloc(uSize); }
void * sec_calloc(size_t cNum, size_t uSize) {
  return calloc(cNum, uSize); }
void * sec_realloc(void *pData, size_t uNewSize) {
  return realloc(pData, uNewSize); }
void sec_free(void *pData) { free(pData); }
void sec_heap_compact(void) { }
int sec_heap_mode(int iMode) {
  UNREFERENCED_PARAMETER(iMode); return 0; }
#ifdef SEC_DEBUG
int sec_heap_status(PSEC_HEAP_INFORMATION pshi) {
  UNREFERENCED_PARAMETER(pshi); return 0; }
#endif

#endif
